"""
Module storing a BigDFT inputfile as an AiiDA Dict
"""

from aiida.orm import Dict


class BigDFTParameters(Dict):  # pylint: disable=too-many-ancestors
    """
    Command line options for a BigDFT calculation.

    This class represents the yaml inputfile which will be passed to the executable.
    """

    # pylint: disable=redefined-builtin
    def __init__(self, dict: dict = None, **kwargs):
        """
        Constructor for the data class
        Usage: ``BigDFTParameters(**parameters)``
        :param parameters_dict: dictionary with commandline parameters
        :param type parameters_dict: dict
        """
        dict = self.validate(dict)
        super().__init__(dict=dict, **kwargs)

    def validate(self, parameters_dict: dict):
        """Validate command line options.
        Uses the voluptuous package for validation. Find out about allowed keys using::
            print(DiffParameters).schema.schema
        :param parameters_dict: dictionary with commandline parameters
        :param type parameters_dict: dict
        :returns: validated dictionary
        """
        if parameters_dict is None:
            parameters_dict = {}

        return parameters_dict

    def __str__(self):
        """String representation of node.
        Append values of dictionary to usual representation. E.g.::
            uuid: b416cbee-24e8-47a8-8c11-6d668770158b (pk: 590)
            {'ignore-case': True}
        """
        string = super().__str__()
        string += "\n" + str(self.get_dict())
        return string



"""
Calculations provided by aiida_bigdft.

Register calculations via the "aiida.calculations" entry point in setup.json.
"""
import os

import yaml

from aiida.common import datastructures
from aiida.engine import CalcJob
import aiida.orm
from aiida.orm import List, SinglefileData, to_aiida_type

from aiida_bigdft.data.BigDFTFile import BigDFTFile, BigDFTLogfile
from aiida_bigdft.data.BigDFTParameters import BigDFTParameters

_DEFAULT_PARAMS = {
    "dft": {"ixc": "LDA", "itermax": "5"},
    "output": {"orbitals": "binary"},
}


class BigDFTCalculation(CalcJob):
    """
    AiiDA plugin wrapping a BigDFT calculation

    requires a valid BigDFT install and a copy of `bigdft.py` on the target machine.
    """

    _posinp = "posinp.xyz"
    _inpfile = "input.yaml"
    _logfile = "log.yaml"
    _timefile = "time.yaml"

    _defaults = {"jobname": "bigdft_calculation"}

    @classmethod
    def define(cls, spec):
        """Define inputs and outputs of the calculation."""
        super().define(spec)

        # set default values for AiiDA options
        spec.inputs["metadata"]["options"]["resources"].default = {
            "num_machines": 1,
            "num_mpiprocs_per_machine": 1,
            "tot_num_mpiprocs": 1,
        }
        spec.inputs["metadata"]["options"]["parser_name"].default = "bigdft"

        # inputs
        spec.input(
            "structure",
            valid_type=aiida.orm.StructureData,
            help="Input structure (AiiDA format)",
        )
        spec.input(
            "parameters",
            valid_type=BigDFTParameters,
            default=lambda: BigDFTParameters(_DEFAULT_PARAMS),
            help="BigDFT Inputfile parameters, as Dict",
        )

        spec.input(
            "structure_fname",
            valid_type=aiida.orm.Str,
            default=lambda: aiida.orm.Str("structure.json"),
            help="Name override for structure file",
        )
        spec.input(
            "params_fname",
            valid_type=aiida.orm.Str,
            default=lambda: aiida.orm.Str("input.yaml"),
            help="Name override for parameters file",
        )

        spec.input(
            "metadata.options.bigdft_mpirun",
            valid_type=str,
            default="UNSET",
            help="Override for bigdft mpirun, defaults to computer.mpirun_command",
        )

        spec.input(
            "metadata.options.jobname",
            valid_type=str,
            default=BigDFTCalculation._defaults["jobname"],
            help="Scheduler jobname",
        )

        spec.input(
            "extra_files_send",
            valid_type=List,
            serializer=to_aiida_type,
            default=lambda: List(),
            help="Extra files to send with calculation",
        )
        spec.input(
            "extra_files_recv",
            valid_type=List,
            serializer=to_aiida_type,
            default=lambda: List(),
            help="Extra files to retrieve from calculation",
        )

        # outputs
        spec.output(
            "logfile", valid_type=BigDFTLogfile, help="BigDFT calculation Logfile"
        )
        spec.output(
            "timefile", valid_type=BigDFTFile, help="BigDFT calculation time log"
        )
        spec.output(
            "energy",
            valid_type=aiida.orm.Float,
            help="Final energy estimate taken from logfile",
        )
        spec.output(
            "ttotal",
            valid_type=aiida.orm.Float,
            help="Estimated total run time (excluding queue)",
        )

        spec.exit_code(
            100,
            "ERROR_MISSING_OUTPUT_FILES",
            message="Calculation did not produce all expected output files.",
        )
        spec.exit_code(
            101,
            "ERROR_PARSING_FAILED",
            message="Calculation did not produce all expected output files.",
        )
        spec.exit_code(
            400,
            "ERROR_OUT_OF_WALLTIME",
            message="Calculation did not finish because of a walltime issue.",
        )
        spec.exit_code(
            401,
            "ERROR_OUT_OF_MEMORY",
            message="Calculation did not finish because of memory limit",
        )

    def prepare_for_submission(self, folder):
        """
        Create input files.

        :param folder: an `aiida.common.folders.Folder` where the plugin should temporarily place all files
            needed by the calculation.
        :return: `aiida.common.datastructures.CalcInfo` instance
        """
        struct_fname = self.inputs.structure_fname.value
        with folder.open(struct_fname, "w") as o:
            self.inputs.structure.get_ase().write(o)

        # dump params
        params_fname = str(self.inputs.params_fname.value)
        with folder.open(params_fname, "w") as o:
            yaml.dump(self.inputs.parameters.get_dict(), o)

        # submission parameters
        jobname = self.metadata.options.jobname
        sub_params_file = self.dump_submission_parameters(folder)

        # aiida calcinfo setup
        codeinfo = datastructures.CodeInfo()

        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.cmdline_params = [
            "--structure",
            self.inputs.structure_fname.value,
            "--parameters",
            self.inputs.params_fname.value,
            "--submission",
            sub_params_file,
        ]

        # Prepare a `CalcInfo` to be returned to the engine
        calcinfo = datastructures.CalcInfo()
        calcinfo.codes_info = [codeinfo]

        list_send = []
        for file in self.inputs.extra_files_send.get_list():

            if not os.path.isabs(file):
                raise ValueError(f"Path to file {file} is not absolute!")

            tmp = SinglefileData(os.path.join(file))
            list_send.append((tmp.uuid, tmp.filename, tmp.filename))
            tmp.store()

        calcinfo.local_copy_list = list_send

        list_recv = [
            (f"log-{jobname}.yaml", ".", 0),
            (f"./data-{jobname}/time-{jobname}.yaml", ".", 0),
            ("./debug/bigdft-err*", ".", 2),
        ]

        for file in self.inputs.extra_files_recv.get_list():
            list_recv.append((file, ".", 0))

        calcinfo.retrieve_list = list_recv

        return calcinfo

    def dump_submission_parameters(self, folder):
        """
        Dump the internal resource request values to the --submission file that bigdft.py is expecting

        """
        sub_params_file = "submission_parameters.yaml"
        sub_params = {"jobname": self.metadata.options.jobname}

        sub_params["OMP"] = self.metadata.options.resources.get(
            "num_cores_per_mpiproc", None
        )
        sub_params["mpi"] = self.metadata.options.resources.get(
            "tot_num_mpiprocs", None
        )
        sub_params["nodes"] = self.metadata.options.resources.get("num_machines", None)

        sub_params["aiida_resources"] = self.metadata.options.resources

        computer = self.node.computer
        user = aiida.orm.User.objects.get_default()

        mpirun = self.metadata.options.bigdft_mpirun
        if mpirun != "UNSET":
            sub_params["mpirun"] = mpirun
        else:
            sub_params["mpirun"] = computer.get_mpirun_command()[0]

        sub_params["connection"] = computer.get_authinfo(user).get_auth_params()

        # This actually updates the computer mpirun command permanently
        # self.node.computer.set_mpirun_command([])

        with folder.open(sub_params_file, "w") as o:
            yaml.dump(sub_params, o)

        return sub_params_file



#!/usr/bin/env python
"""
Testing class to mimic bigdft.py within the CI. Any validation is to be performed here.

"""
import os

import click
import yaml


@click.command()
@click.option("--structure", help="path to structure json file")
@click.option("--parameters", help="yaml dumped dft parameters")
@click.option("--submission", help="extra submission parameters")
def run(
    structure: str = None,  # pylint: disable=unused-argument
    parameters: str = None,  # pylint: disable=unused-argument
    submission: str = None,
) -> dict:
    """
    Run the calculation. Requires three file path inputs:

    Args:
        structure (str):
            path to the serialised ASE json file
        parameters (str):
            path to the serialised BigDFTParameters yaml file
        submission (str):
            path to the serialised submission yaml file

    Returns:
        BigDFT.Logfile
    """
    ########################
    ### submission param ###
    ########################
    params_sub = {}
    if submission is not None:
        with open(submission, encoding="utf8") as o:
            params_sub = yaml.safe_load(o)

    jobname = params_sub["jobname"]

    with open(f"log-{jobname}.yaml", "w+", encoding="utf8") as o:
        yaml.dump({}, o)

    os.makedirs(f"data-{jobname}")
    with open(f"data-{jobname}/time-{jobname}.yaml", "w+", encoding="utf8") as o:
        yaml.dump({}, o)

    payload = {"pass": True}

    with open("testing_info.yaml", "w+", encoding="utf8") as o:
        yaml.dump(payload, o)


if __name__ == "__main__":
    run()



""" Tests for calculations."""
import os

import yaml
from aiida.engine import run
from aiida.orm import StructureData
from aiida.plugins import CalculationFactory


def test_process(bigdft_code):
    """Test running a calculation
    note this does not test that the expected outputs are created of output parsing"""

    # set up calculation
    cell = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
    s = StructureData(cell=cell)
    s.append_atom(position=(0, 0, 0), symbols="C")

    inputs = {
        "code": bigdft_code,
        "structure": s,
        "metadata": {
            "options": {
                "jobname": "Mono_Carbon",
                "max_wallclock_seconds": 60,
                "resources": {
                    "num_cores_per_mpiproc": 1,
                    "num_mpiprocs_per_machine": 1,
                    "num_machines": 1,
                },
                "withmpi": False,
            }
        },
    }

    result = run(CalculationFactory("bigdft"), **inputs)

    rpath = result["remote_folder"].get_remote_path()
    try:
        with open(os.path.join(rpath, "testing_info.yaml")) as o:
            result = yaml.safe_load(o)
    except FileNotFoundError:
        raise RuntimeError(
            f"Test did not output expected result file within workdir {rpath}"
        )

    assert result["pass"]



"""
Minimised logger class to handle debugging once the daemon takes control

"""
import os.path
from datetime import datetime


class MiniLogger:
    """
    Miniature logger class

    :param path: path to intended file
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, path: str):

        dpath = os.path.split(path)[0]
        if not os.path.isdir(dpath):
            os.makedirs(dpath)

        self._path = path

        self.debug("minilogger class init", wipe=True)

    def debug(self, msg: str, wipe: bool = False):
        """
        Append debug string `msg` to file (unless wipe=True, then create a fresh file)

        :param msg: string to write
        :param wipe: clears the file before writing if True
        """
        mode = "w+" if wipe else "a"
        timestr = datetime.now().strftime("%H:%M:%S")
        with open(self._path, mode, encoding="utf8") as o:
            o.write(f"[{timestr}] {msg}\n")



"""pytest fixtures for simplified testing."""
import pytest

from aiida_bigdft import helpers

pytest_plugins = ["aiida.manage.tests.pytest_fixtures"]


@pytest.fixture(scope="function", autouse=True)
def clear_database_auto(clear_database):  # pylint: disable=unused-argument
    """Automatically clear database in between tests."""


@pytest.fixture(scope="function")
def bigdft_code():
    """Get a bigdft code."""
    computer = helpers.get_computer()
    code = helpers.get_code(entry_point="bigdft", computer=computer, force_create=True)
    return code



#!/usr/bin/env python
"""
Executable script to create and launch a BigDFT calculation from the inputs.

Requires:
- structure.json (from ase.write())
- input.yaml

And optionally (advised):
- submission.yaml

call with

./bigdft.py --structure structure.json --parameters input.yaml --submission submission.yaml
"""
import os.path

from BigDFT.Calculators import SystemCalculator
from BigDFT.Inputfiles import Inputfile
from BigDFT.Interop.ASEInterop import ase_to_bigdft
from BigDFT.Logfiles import Logfile
from BigDFT.Systems import System
import ase.io
import click
import yaml


@click.command()
@click.option("--structure", help="path to structure json file")
@click.option("--parameters", help="yaml dumped dft parameters")
@click.option("--submission", help="extra submission parameters")
def run(
    structure: str = None, parameters: str = None, submission: str = None
) -> Logfile:
    """
    Run the calculation. Requires three file path inputs:

    Args:
        structure (str):
            path to the serialised ASE json file
        parameters (str):
            path to the serialised BigDFTParameters yaml file
        submission (str):
            path to the serialised submission yaml file

    Returns:
        BigDFT.Logfile
    """
    ########################
    ### submission param ###
    ########################
    params_sub = {}
    if submission is not None:
        with open(submission, encoding="utf8") as o:
            params_sub = yaml.safe_load(o)

    calc_args = {}

    if "mpirun" in params_sub:
        calc_args["mpi_run"] = params_sub["mpirun"]

    ########################
    ####    structure    ###
    ########################
    if structure is None:
        structure = "structure.json"
    structure = os.path.abspath(structure)
    struct_ase = ase.io.read(structure)

    # structure proper
    frag = ase_to_bigdft(struct_ase)

    sys = System()
    sys["FRA:1"] = frag

    ########################
    ####   calc params   ###
    ########################
    if parameters is None:
        parameters = "input.yaml"
    with open(parameters, encoding="utf8") as o:
        parameters = yaml.safe_load(o)
    inp = Inputfile(parameters)

    ########################
    ###    calculation   ###
    ########################
    code = SystemCalculator(**calc_args)
    log = code.run(
        input=inp, sys=sys, name=params_sub.get("jobname", "bigdft_calculation")
    )

    return log


if __name__ == "__main__":
    run()



# !/usr/bin/env python
"""Run a test calculation on localhost.
Usage: ./example_01.py
"""
import click

from aiida import cmdline
from aiida.engine import submit
from aiida.orm import StructureData

from aiida_bigdft import helpers
from aiida_bigdft.calculations import BigDFTCalculation
from aiida_bigdft.data import BigDFTParameters


def test_run(code):
    """Run a calculation on the localhost computer.
    Uses test helpers to create AiiDA Code on the fly.
    """
    if not code:
        # get code
        computer = helpers.get_computer()
        code = helpers.get_code(entry_point="bigdft", computer=computer)

    alat = 4  # angstrom
    cell = [
        [
            alat,
            0,
            0,
        ],
        [
            0,
            alat,
            0,
        ],
        [
            0,
            0,
            alat,
        ],
    ]

    s = StructureData(cell=cell)
    s.append_atom(position=(alat / 2, alat / 2, alat / 2), symbols="Ti")
    s.append_atom(position=(alat / 2, alat / 2, 0), symbols="O")
    s.append_atom(position=(alat / 2, 0, alat / 2), symbols="O")

    print(f"running code {code}")
    inputs = {
        "code": code,
        "structure": s,
        "metadata": {
            "options": {
                "jobname": "TiO2",
                "max_wallclock_seconds": 3600,
                "queue_name": "short",
                "resources": {
                    # "num_cores_per_machine": 16,
                    "num_cores_per_mpiproc": 1,
                    "num_mpiprocs_per_machine": 1,
                    "num_machines": 1,
                },
                "withmpi": False,
            }
        },
    }

    bigdft_parameters = {}
    bigdft_parameters["dft"] = {"ixc": "LDA", "itermax": "5"}
    bigdft_parameters["output"] = {"orbitals": "binary"}

    inputs["parameters"] = BigDFTParameters(bigdft_parameters)

    result = submit(BigDFTCalculation, **inputs)

    return result


@click.command()
@cmdline.utils.decorators.with_dbenv()
@cmdline.params.options.CODE()
def cli(code):
    """Run example.
    Example usage: $ ./example_01.py --code diff@localhost
    Alternative (creates diff@localhost-test code): $ ./example_01.py
    Help: $ ./example_01.py --help
    """
    test_run(code)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



from aiida.common import AttributeDict
from aiida.engine import (
    BaseRestartWorkChain,
    ProcessHandlerReport,
    process_handler,
    while_,
)

from aiida_bigdft.calculations import BigDFTCalculation


class BigDFTBaseWorkChain(BaseRestartWorkChain):
    """Base workchain for running a BigDFT Calculation"""

    _process_class = BigDFTCalculation
    _workchain_namespace = "BigDFT"

    @classmethod
    def define(cls, spec):
        super().define(spec)

        spec.expose_inputs(
            BigDFTCalculation, namespace=BigDFTBaseWorkChain._workchain_namespace
        )
        spec.expose_outputs(BigDFTCalculation)

        spec.outline(
            cls.setup,
            while_(cls.should_run_process)(cls.run_process, cls.inspect_process),
            cls.results,
        )
        spec.exit_code(
            300,
            "ERROR_UNRECOVERABLE_FAILURE",
            message="The calculation encountered an unrecoverable error",
        )

    def setup(self):
        """Call the `setup` of the `BaseRestartWorkChain` and then create the inputs dictionary in `self.ctx.inputs`.

        This `self.ctx.inputs` dictionary will be used by the `BaseRestartWorkChain` to submit the calculations in the
        internal loop.
        """
        super().setup()
        self.ctx.restart_calc = None
        self.ctx.inputs = AttributeDict(
            self.exposed_inputs(
                BigDFTCalculation, BigDFTBaseWorkChain._workchain_namespace
            )
        )

    def report_error_handled(self, calculation, action):
        """Report an action taken for a calculation that has failed.

        This should be called in a registered error handler if its condition is met and an action was taken.

        :param calculation: the failed calculation node
        :param action: a string message with the action taken
        """
        self.report(
            f"{calculation.process_label}<{calculation.pk}> "
            f"failed with exit status {calculation.exit_status}: "
            f"{calculation.exit_message}"
        )
        self.report(f"Action taken: {action}")

    @process_handler(priority=600)
    def handle_unrecoverable_failure(self, node):
        """Handle calculations with an exit status below 400 which are unrecoverable, so abort the work chain."""
        if node.is_failed and node.exit_status < 400:
            self.report_error_handled(node, "unrecoverable error, aborting...")
            return ProcessHandlerReport(
                True, self.exit_codes.ERROR_UNRECOVERABLE_FAILURE
            )



#
# Sphinx configuration for aiida-bigdft
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys
import time

import aiida_bigdft

# imports required for docs/source/reference/api/public.rst
from aiida import (  # pylint: disable=unused-import
    cmdline,
    common,
    engine,
    manage,
    orm,
    parsers,
    plugins,
    schedulers,
    tools,
    transports,
)
from aiida.cmdline.params import arguments, options  # pylint: disable=unused-import

# -- General configuration ------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinxcontrib.contentui",
    "aiida.sphinxext",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "aiida": ("https://aiida.readthedocs.io/projects/aiida-core/en/latest", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
# ~ master_doc = 'index'
master_doc = "index"

# General information about the project.
project = "aiida-bigdft"
copyright_first_year = "2023"
copyright_owners = "Louis Beal"

current_year = str(time.localtime().tm_year)
copyright_year_string = (
    current_year
    if current_year == copyright_first_year
    else f"{copyright_first_year}-{current_year}"
)
# pylint: disable=redefined-builtin
copyright = f"{copyright_year_string}, {copyright_owners}. All rights reserved"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = aiida_bigdft.__version__
# The short X.Y version.
version = ".".join(release.split(".")[:2])

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# exclude_patterns = ['doc.rst']
# ~ exclude_patterns = ['index.rst']

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output ----------------------------------------------

html_theme = "furo"
html_logo = "images/AiiDA_transparent_logo.png"
html_title = f"aiida-bigdft v{release}"
html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# ~ html_theme_path = ["."]

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = "images/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# ~ html_show_copyright = False

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
html_use_opensearch = "https://aiida-bigdft.readthedocs.io"

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'hu', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'ru', 'sv', 'tr'
html_search_language = "en"

# Warnings to ignore when using the -n (nitpicky) option
# We should ignore any python built-in exception, for instance
nitpick_ignore = [
    ("py:class", "Logger"),
]


def run_apidoc(_):
    """Runs sphinx-apidoc when building the documentation.

    Needs to be done in conf.py in order to include the APIdoc in the
    build on readthedocs.

    See also https://github.com/rtfd/readthedocs.org/issues/1139
    """
    source_dir = os.path.abspath(os.path.dirname(__file__))
    apidoc_dir = os.path.join(source_dir, "apidoc")
    package_dir = os.path.join(source_dir, os.pardir, os.pardir, "aiida_bigdft")

    # In #1139, they suggest the route below, but this ended up
    # calling sphinx-build, not sphinx-apidoc
    # from sphinx.apidoc import main
    # main([None, '-e', '-o', apidoc_dir, package_dir, '--force'])

    import subprocess

    cmd_path = "sphinx-apidoc"
    if hasattr(sys, "real_prefix"):  # Check to see if we are in a virtualenv
        # If we are, assemble the path manually
        cmd_path = os.path.abspath(os.path.join(sys.prefix, "bin", "sphinx-apidoc"))

    options = [
        "-o",
        apidoc_dir,
        package_dir,
        "--private",
        "--force",
        "--no-toc",
    ]

    # See https://stackoverflow.com/a/30144019
    env = os.environ.copy()
    env[
        "SPHINX_APIDOC_OPTIONS"
    ] = "members,special-members,private-members,undoc-members,show-inheritance"
    subprocess.check_call([cmd_path] + options, env=env)


def setup(app):
    app.connect("builder-inited", run_apidoc)



# !/usr/bin/env python
"""Run a test calculation on localhost.
Usage: ./example_01.py
"""
import os
import click

from aiida import cmdline
from aiida.engine import submit
from aiida.orm import StructureData

from aiida_bigdft import helpers
from aiida_bigdft.calculations import BigDFTCalculation
from aiida_bigdft.data import BigDFTParameters


def test_run(code):
    """Run a calculation on the localhost computer.
    Uses test helpers to create AiiDA Code on the fly.
    """
    if not code:
        # get code
        computer = helpers.get_computer()
        code = helpers.get_code(entry_point="bigdft", computer=computer)

    testfile = os.path.join(os.getcwd(), "test.txt")

    with open(testfile, "w+", encoding="utf8") as o:
        # use `verdi calcjob outputcat <pk> test.txt` to verify the existence of this file
        o.write("I am a test\n")

    cell = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
    s = StructureData(cell=cell)
    s.append_atom(position=(0, 0, 0), symbols="C")

    print(f"running code {code}")
    inputs = {
        "code": code,
        "structure": s,
        "extra_files_send": [testfile],
        "metadata": {
            "options": {
                "jobname": "Mono_Carbon",
                "max_wallclock_seconds": 3600,
                "withmpi": False,
            }
        },
    }

    bigdft_parameters = {}
    bigdft_parameters["dft"] = {"ixc": "LDA", "itermax": "5"}
    bigdft_parameters["output"] = {"orbitals": "binary"}

    inputs["parameters"] = BigDFTParameters(bigdft_parameters)

    result = submit(BigDFTCalculation, **inputs)

    return result


@click.command()
@cmdline.utils.decorators.with_dbenv()
@cmdline.params.options.CODE()
def cli(code):
    """Run example.
    Example usage: $ ./example_01.py --code diff@localhost
    Alternative (creates diff@localhost-test code): $ ./example_01.py
    Help: $ ./example_01.py --help
    """
    test_run(code)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



"""
Module for adding extra BigDFT functionality to AiiDA's base SinglefileData
"""

import os
from typing import Union

from BigDFT.Logfiles import Logfile
import yaml

from aiida.orm import SinglefileData


class BigDFTFile(SinglefileData):
    """
    Wrapper class for a BigDFT yaml format file as SinglefileData
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._content = self._open()

    def _open(self):
        """
        Attempts to open the stored file, returning an empty dict on failure
        """
        try:
            with self.open() as o:  # pylint: disable=not-context-manager
                content = yaml.safe_load(o)
                if content is None:
                    return {}
                return content
        except FileNotFoundError:
            self.logger.warning(f"file {self.filename} could not be opened!")
            return {}

    @property
    def content(self):
        """
        Attempts to return file content from cache, loading otherwise
        """
        try:
            return self._content
        except AttributeError:
            self._content = self._open()
            return self._content

    def dump_file(self, path=None):
        """
        Dump the stored file to `path`
        defaults to cwd + filename if not provided
        """
        path = path or os.path.join(os.getcwd(), self.filename)

        with self.open() as inp:  # pylint: disable=not-context-manager
            with open(path, "w+", encoding="utf8") as out:
                out.write(inp.read())


class BigDFTLogfile(BigDFTFile):
    """
    Specialised class for wrapping a BigDFT Logfile class as SinglefileData
    """

    @property
    def logfile(self) -> Union[Logfile, None]:
        """
        Create and return the BigDFT Logfile object
        """
        if not hasattr(self, "_logfile"):
            try:
                self._logfile = Logfile(dictionary=self.content)
            except Exception as e:
                self._logfile_generate_error = e
                self._logfile = None

        return self._logfile



"""
Parsers provided by aiida_bigdft.

Register parsers via the "aiida.parsers" entry point in setup.json.
"""
import os
import re

from aiida.common import exceptions
from aiida.engine import ExitCode
import aiida.orm
from aiida.parsers.parser import Parser

from aiida_bigdft.calculations import BigDFTCalculation
from aiida_bigdft.data.BigDFTFile import BigDFTFile, BigDFTLogfile
from aiida_bigdft.utils.MiniLogger import MiniLogger


class BigDFTParser(Parser):
    """
    Parser class for parsing output of calculation.
    """

    def __init__(self, node):
        """
        Initialize Parser instance

        Checks that the ProcessNode being passed was produced by a DiffCalculation.

        :param node: ProcessNode of calculation
        :param type node: :class:`aiida.orm.nodes.process.process.ProcessNode`
        """
        super().__init__(node)
        if not issubclass(node.process_class, BigDFTCalculation):
            raise exceptions.ParsingError("Can only parse DiffCalculation")

    def parse_stderr(self, inputfile):
        """Parse the stderr file to get common errors, such as OOM or timeout.

        :param inputfile: stderr file
        :returns: exit code in case of an error, None otherwise
        """
        timeout_messages = {
            "DUE TO TIME LIMIT",  # slurm
            "exceeded hard wallclock time",  # UGE
            "TERM_RUNLIMIT: job killed",  # LFS
            "walltime .* exceeded limit",  # PBS/Torque
        }

        oom_messages = {
            "[oO]ut [oO]f [mM]emory",
            "oom-kill",  # generic OOM messages
            "Exceeded .* memory limit",  # slurm
            "exceeds job hard limit .*mem.* of queue",  # UGE
            "TERM_MEMLIMIT: job killed after reaching LSF memory usage limit",  # LFS
            "mem .* exceeded limit",  # PBS/Torque
        }
        for message in timeout_messages:
            if re.search(message, inputfile):
                return self.exit_codes.ERROR_OUT_OF_WALLTIME
        for message in oom_messages:
            if re.search(message, inputfile):
                return self.exit_codes.ERROR_OUT_OF_MEMORY
        return

    def parse(self, **kwargs):
        """
        Parse outputs, store results in database.

        :returns: an exit code, if parsing fails (or nothing if parsing succeeds)
        """

        exitcode = ExitCode(0)

        stderr = self.node.get_scheduler_stderr()
        if stderr:
            exitcode = self.parse_stderr(stderr)
            if exitcode:
                self.logger.error("Error in stderr: " + exitcode.message)

        jobname = BigDFTCalculation._defaults["jobname"]

        if "jobname" in self.node.get_options():
            jobname = self.node.get_options()["jobname"]

        metadata = self.node.get_metadata_inputs()["metadata"]
        if "jobname" in metadata.get("options", {}):
            jobname = metadata["options"]["jobname"]

        output_filename = f"log-{jobname}.yaml"
        time_filename = f"time-{jobname}.yaml"

        files_retrieved = self.retrieved.list_object_names()
        files_expected = [output_filename, time_filename]

        files_expected += self.node.inputs.extra_files_recv.get_list()

        # Note: set(A) <= set(B) checks whether A is a subset of B
        if not set(files_expected) <= set(files_retrieved):
            self.logger.error(
                f"Found files '{files_retrieved}', expected to find '{files_expected}'"
            )
            return self.exit_codes.ERROR_MISSING_OUTPUT_FILES

        logfile = self.parse_file(output_filename, "logfile", exitcode)
        timefile = self.parse_file(time_filename, "timefile", exitcode)

        self.out("logfile", logfile)
        self.out("timefile", timefile)

        energy = 0.0
        if logfile.logfile is not None:
            energy = logfile.logfile.energy

        self.out("energy", aiida.orm.Float(energy))

        ttotal = timefile.content.get("SUMMARY", None)
        if ttotal is not None:
            ttotal = ttotal.get("Total", [-1.0])

            self.out("ttotal", aiida.orm.Float(ttotal[-1]))

        else:
            self.out("ttotal", aiida.orm.Float(-1.0))

        return exitcode

    def parse_file(self, output_filename, name, exitcode):
        """
        Parse a retrieved file into a BigDFTFile object
        """

        # add output file
        self.logger.info(f"Parsing '{output_filename}'")
        try:
            with open(output_filename, "w+", encoding="utf8") as tmp:
                tmp.write(self.retrieved.get_object_content(output_filename))
                if name == "logfile":
                    output = BigDFTLogfile(os.path.join(os.getcwd(), output_filename))
                else:
                    output = BigDFTFile(os.path.join(os.getcwd(), output_filename))

        except ValueError:
            self.logger.error(f"Impossible to parse {name} {output_filename}")
            if (
                not exitcode
            ):  # if we already have OOW or OOM, failure here will be handled later
                return self.exit_codes.ERROR_PARSING_FAILED
        try:
            output.store()
            self.logger.info(f"Successfully parsed {name} '{output_filename}'")
        except exceptions.ValidationError:
            self.logger.info(
                f"Impossible to store {name} - ignoring '{output_filename}'"
            )
            if (
                not exitcode
            ):  # if we already have OOW or OOM, failure here will be handled later
                return self.exit_codes.ERROR_PARSING_FAILED

        return output
